from dune.perftool.ufl.modified_terminals import Restriction

from dune.perftool.options import get_option

from dune.perftool.pdelab.argument import name_coefficientcontainer

from dune.perftool.generation import (class_member,
                                      domain,
                                      function_mangler,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      include_file,
                                      initializer_list,
                                      instruction,
                                      loopy_class_member,
                                      preamble,
                                      silenced_warning,
                                      temporary_variable,
                                      valuearg
                                      )
from dune.perftool.loopy.buffer import get_buffer_temporary
from dune.perftool.pdelab.localoperator import (name_domain_field,
                                                lop_template_range_field,
                                                )
from dune.perftool.pdelab.quadrature import quadrature_order
from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

from pytools import Record, product

import pymbolic.primitives as prim
import numpy


class AMatrix(Record):
    def __init__(self, a_matrix, rows, cols):
        Record.__init__(self,
                        a_matrix=a_matrix,
                        rows=rows,
                        cols=cols,
                        )

    def __hash__(self):
        return hash((self.a_matrix, self.rows, self.cols))


def quadrature_points_per_direction():
    # TODO use quadrature order from dune.perftool.pdelab.quadrature
    # as soon as we have a generic implementation

    # Quadrature order
    q = quadrature_order()

    # Quadrature points in per direction
    nb_qp = q // 2 + 1

    return nb_qp


def polynomial_degree():
    form = get_global_context_value("formdata").preprocessed_form
    return form.coefficients()[0].ufl_element()._degree


def basis_functions_per_direction():
    return polynomial_degree() + 1


def define_oned_quadrature_weights(name):
    loopy_class_member(name,
                       dtype=numpy.float64,
                       classtag="operator",
                       shape=(quadrature_points_per_direction(),),
                       )


def name_oned_quadrature_weights():
    name = "qw"
    define_oned_quadrature_weights(name)
    return name


def define_oned_quadrature_points(name):
    loopy_class_member(name,
                       dtype=numpy.float64,
                       classtag="operator",
                       shape=(quadrature_points_per_direction(),),
                       )


def name_oned_quadrature_points():
    name = "qp"
    define_oned_quadrature_points(name)
    return name


@class_member(classtag="operator")
def typedef_polynomials(name):
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    degree = polynomial_degree()

    include_file("dune/pdelab/finiteelementmap/qkdg.hh", filetag="operatorfile")

    # TODO: make switching polynomials possible
    # return "using {} = Dune::QkStuff::GaussLobattoLagrangePolynomials<{}, {}, {}>;".format(name,
    #                                                                                        domain_field,
    #                                                                                        range_field,
    #                                                                                        degree)
    return "using {} = Dune::QkStuff::EquidistantLagrangePolynomials<{}, {}, {}>;".format(name,
                                                                                          domain_field,
                                                                                          range_field,
                                                                                          degree)


def type_polynomials():
    name = "Polynomials1D"
    typedef_polynomials(name)
    return name


@class_member(classtag="operator")
def define_polynomials(name):
    polynomials_type = type_polynomials()
    return "{} {};".format(polynomials_type, name)


def name_polynomials():
    name = "poly"
    define_polynomials(name)
    return name


@preamble(kernel="operator")
def sort_quadrature_points_weights():
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    number_qp = quadrature_points_per_direction()
    qp = name_oned_quadrature_points()
    qw = name_oned_quadrature_weights()
    include_file("dune/perftool/sumfact/onedquadrature.hh", filetag="operatorfile")
    return "onedQuadraturePointsWeights<{}, {}, {}>({}, {});".format(range_field, domain_field, number_qp, qp, qw)


@iname(kernel="operator")
def theta_iname(name, bound):
    name = "{}_{}".format(name, bound)
    domain(name, bound, kernel="operator")
    return name


class PolynomialLookup(FunctionIdentifier):
    def __init__(self, pol, derivative):
        self.pol = pol
        self.derivative = derivative

    def __getinitargs__(self):
        return (self.pol, self.derivative)

    @property
    def name(self):
        return "{}.{}".format(self.pol, "dp" if self.derivative else "p")


@function_mangler
def polynomial_lookup_mangler(target, func, dtypes):
    if isinstance(func, PolynomialLookup):
        return CallMangleInfo(func.name, (NumpyType(numpy.float64),), (NumpyType(numpy.int32), NumpyType(numpy.float64)))


def define_theta(name, shape, transpose, derivative):
    sort_quadrature_points_weights()
    polynomials = name_polynomials()
    qp = name_oned_quadrature_points()

    loopy_class_member(name,
                       dtype=numpy.float64,
                       shape=shape,
                       classtag="operator",
                       dim_tags="f,f",
                       managed=True,
                       )

    # TODO Enforce the alignment here!

    i = theta_iname("i", shape[0])
    j = theta_iname("j", shape[1])

    if transpose:
        args = (prim.Variable(i), prim.Subscript(prim.Variable(qp), (prim.Variable(j),)))
    else:
        args = (prim.Variable(j), prim.Subscript(prim.Variable(qp), (prim.Variable(i),)))

    instruction(assignee=prim.Subscript(prim.Variable(name), (prim.Variable(i), prim.Variable(j))),
                expression=prim.Call(PolynomialLookup(polynomials, derivative), args),
                kernel="operator",
                )


def name_theta():
    name = "Theta"
    shape = (quadrature_points_per_direction(), basis_functions_per_direction())
    globalarg(name, shape=shape, dtype=numpy.float64, dim_tags="f,f")
    define_theta(name, shape, False, False)
    return name


def name_theta_transposed():
    name = "ThetaT"
    shape = (basis_functions_per_direction(), quadrature_points_per_direction())
    globalarg(name, shape=shape, dtype=numpy.float64, dim_tags="f,f")
    define_theta(name, shape, True, False)
    return name


def name_dtheta():
    name = "dTheta"
    shape = (quadrature_points_per_direction(), basis_functions_per_direction())
    globalarg(name, shape=shape, dtype=numpy.float64, dim_tags="f,f")
    define_theta(name, shape, False, True)
    return name


def name_dtheta_transposed():
    name = "dThetaT"
    shape = (basis_functions_per_direction(), quadrature_points_per_direction())
    globalarg(name, shape=shape, dtype=numpy.float64, dim_tags="f,f")
    define_theta(name, shape, True, True)
    return name
